@@ -47,7 +47,6 @@
 #include "txrx.h"
 #include "list.h"
 #include "error.h"
-#include "encoding.h"
 #include <libvchan.h>
 
 #define SOCKET_ADDRESS  "/var/run/xf86-qubes-socket"
@@ -77,8 +76,7 @@ struct _global_handles {
     Atom tray_opcode;      /* Atom: _NET_SYSTEM_TRAY_MESSAGE_OPCODE */
     Atom xembed_info;      /* Atom: _XEMBED_INFO */
     Atom utf8_string_atom; /* Atom: UTF8_STRING */
-    Atom wm_state;         /* Atom: WM_STATE */
-    Atom net_wm_state;     /* Atom: _NET_WM_STATE */
+    Atom wm_state;         /* Atom: _NET_WM_STATE */
     Atom wm_state_fullscreen; /* Atom: _NET_WM_STATE_FULLSCREEN */
     Atom wm_state_demands_attention; /* Atom: _NET_WM_STATE_DEMANDS_ATTENTION */
     Atom wm_take_focus;    /* Atom: WM_TAKE_FOCUS */
@@ -93,15 +91,15 @@ struct _global_handles {
     int sync_all_modifiers;
     int composite_redirect_automatic;
     pid_t x_pid;
+    uint32_t domid;
 };
 
 struct window_data {
     int is_docked; /* is it docked icon window */
     XID embeder;   /* for docked icon points embeder window */
     int input_hint; /* the window should get input focus - False=Never */
-    int support_delete_window;
     int support_take_focus;
-    int mfndump_pending; /* send MSG_MFNDUMP at next damage notification */
+    int window_dump_pending; /* send MSG_WINDOW_DUMP at next damage notification */
 };
 
 struct embeder_data {
@@ -113,19 +111,12 @@ struct genlist *embeder_list;
 typedef struct _global_handles Ghandles;
 Ghandles *ghandles_for_vchan_reinitialize;
 
-#define SKIP_NONMANAGED_WINDOW do {                                    \
-    if (!list_lookup(windows_list, window)) {                          \
-        if (g->log_level > 0)                                          \
-            fprintf(stderr, "Skipping unmanaged window 0x%x",          \
-                    (int) window);                                     \
-        return;                                                        \
-    }                                                                  \
-} while (0)
+#define SKIP_NONMANAGED_WINDOW if (!list_lookup(windows_list, window)) return
 
 void send_wmname(Ghandles * g, XID window);
 void send_wmnormalhints(Ghandles * g, XID window, int ignore_fail);
 void send_wmclass(Ghandles * g, XID window, int ignore_fail);
-void send_pixmap_mfns(Ghandles * g, XID window);
+void send_pixmap_grant_refs(Ghandles * g, XID window);
 void retrieve_wmhints(Ghandles * g, XID window, int ignore_fail);
 void retrieve_wmprotocols(Ghandles * g, XID window, int ignore_fail);
 
@@ -142,9 +133,9 @@ void process_xevent_damage(Ghandles * g, XID window,
         return;
 
     wd = l->data;
-    if (wd->mfndump_pending) {
-        send_pixmap_mfns(g, window);
-        wd->mfndump_pending = False;
+    if (wd->window_dump_pending) {
+        send_pixmap_grant_refs(g, window);
+        wd->window_dump_pending = False;
     }
 
     hdr.type = MSG_SHMIMAGE;
@@ -197,9 +188,8 @@ void process_xevent_createnotify(Ghandles * g, XCreateWindowEvent * ev)
     /* Default values for window_data. By default, window should receive InputFocus events */
     wd->is_docked = False;
     wd->input_hint = True;
-    wd->support_delete_window = False;
     wd->support_take_focus = False;
-    wd->mfndump_pending = False;
+    wd->window_dump_pending = False;
     list_insert(windows_list, ev->window, wd);
 
     if (attr.border_width > 0) {
@@ -274,32 +264,33 @@ void read_discarding(int fd, int size)
     }
 }
 
-void send_pixmap_mfns(Ghandles * g, XID window)
+void send_pixmap_grant_refs(Ghandles * g, XID window)
 {
-    struct shm_cmd shmcmd;
     struct msg_hdr hdr;
-    uint32_t *mfnbuf;
-    int ret, rcvd = 0, size;
+    uint8_t *wd_msg_buf;
+    size_t wd_msg_len;
+    size_t rcvd;
+    int ret;
 
     feed_xdriver(g, 'W', (int) window, 0);
-    if (read(g->xserver_fd, &shmcmd, sizeof(shmcmd)) != sizeof(shmcmd)) {
-        perror("unix read shmcmd");
+    if (read(g->xserver_fd, &wd_msg_len, sizeof(wd_msg_len)) != sizeof(wd_msg_len)) {
+        perror("unix read wd_msg_len");
         exit(1);
     }
-    if (shmcmd.num_mfn == 0 || shmcmd.num_mfn > (unsigned)MAX_MFN_COUNT ||
-            shmcmd.width > MAX_WINDOW_WIDTH || shmcmd.height > MAX_WINDOW_HEIGHT) {
-        fprintf(stderr, "got num_mfn=0x%x for window 0x%x (%dx%d)\n",
-                shmcmd.num_mfn, (int) window, shmcmd.width, shmcmd.height);
-        read_discarding(g->xserver_fd,
-                shmcmd.num_mfn * sizeof(*mfnbuf));
+    if (wd_msg_len == 0) {
+        fprintf(stderr, "Failed to get window dump for window 0x%lx\n",
+                window);
         return;
     }
-    size = shmcmd.num_mfn * sizeof(*mfnbuf);
-    mfnbuf = alloca(size);
-    while (rcvd < size) {
-        ret =
-            read(g->xserver_fd, ((char *) mfnbuf) + rcvd,
-                    size - rcvd);
+    wd_msg_buf = alloca(wd_msg_len);
+    if (!wd_msg_buf) {
+        fprintf(stderr, "Failed to allocate memory for window dump 0x%lx\n",
+                window);
+        exit(1);
+    }
+    rcvd = 0;
+    while (rcvd < wd_msg_len) {
+        ret = read(g->xserver_fd, wd_msg_buf + rcvd, wd_msg_len - rcvd);
         if (ret == 0) {
             fprintf(stderr, "unix read EOF\n");
             exit(1);
@@ -310,12 +301,11 @@ void send_pixmap_mfns(Ghandles * g, XID window)
         }
         rcvd += ret;
     }
-    hdr.type = MSG_MFNDUMP;
+    hdr.type = MSG_WINDOW_DUMP;
     hdr.window = window;
-    hdr.untrusted_len = sizeof(shmcmd) + size;
+    hdr.untrusted_len = wd_msg_len;
     write_struct(g->vchan, hdr);
-    write_struct(g->vchan, shmcmd);
-    write_data(g->vchan, (char *) mfnbuf, size);
+    write_data(g->vchan, (char *) wd_msg_buf, wd_msg_len);
 }
 
 /* return 1 on success, 0 otherwise */
@@ -394,7 +384,7 @@ int getwmname_tochar(Ghandles * g, XID window, char *outbuf, int bufsize)
         XFree(text_prop_return.value);
         return 0;
     }
-    strncat(outbuf, list[0], bufsize - 1);
+    strncat(outbuf, list[0], bufsize);
     XFree(text_prop_return.value);
     XFreeStringList(list);
     if (g->log_level > 0)
@@ -426,23 +416,18 @@ void retrieve_wmprotocols(Ghandles * g, XID window, int ignore_fail)
     int i;
     struct genlist *l;
 
-    if (!((l=list_lookup(windows_list, window)) && (l->data))) {
-        fprintf(stderr, "ERROR retrieve_wmprotocols: Window 0x%x data not initialized", (int)window);
-        return;
-    }
-
     if (XGetWMProtocols(g->display, window, &supported_protocols, &nitems) == 1) {
         for (i=0; i < nitems; i++) {
             if (supported_protocols[i] == g->wm_take_focus) {
+                // Retrieve window data and set support_take_focus
+                if (!((l=list_lookup(windows_list, window)) && (l->data))) {
+                    fprintf(stderr, "ERROR retrieve_wmprotocols: Window 0x%x data not initialized", (int)window);
+                    return;
+                }
                 if (g->log_level > 1)
                     fprintf(stderr, "Protocol take_focus supported for Window 0x%x\n", (int)window);
 
                 ((struct window_data*)l->data)->support_take_focus = True;
-            } else if (supported_protocols[i] == g->wmDeleteMessage) {
-                if (g->log_level > 1)
-                    fprintf(stderr, "Protocol delete_window supported for Window 0x%x\n", (int)window);
-
-                ((struct window_data*)l->data)->support_delete_window = True;
             }
         }
     } else {
@@ -582,7 +567,7 @@ void send_window_state(Ghandles * g, XID window)
     struct msg_window_flags flags;
 
     /* FIXME: only first 10 elements are parsed */
-    ret = XGetWindowProperty(g->display, window, g->net_wm_state, 0, 10,
+    ret = XGetWindowProperty(g->display, window, g->wm_state, 0, 10,
             False, XA_ATOM, &act_type, &act_fmt, &nitems, &bytesleft, (unsigned char**)&state_list);
     if (ret != Success)
         return;
@@ -601,7 +586,6 @@ void send_window_state(Ghandles * g, XID window)
 void process_xevent_map(Ghandles * g, XID window)
 {
     XWindowAttributes attr;
-    long new_wm_state[2];
     struct msg_hdr hdr;
     struct msg_map_info map_info;
     Window transient;
@@ -612,7 +596,7 @@ void process_xevent_map(Ghandles * g, XID window)
 
     if (g->log_level > 1)
         fprintf(stderr, "MAP for window 0x%x\n", (int)window);
-    wd->mfndump_pending = True;
+    wd->window_dump_pending = True;
     send_window_state(g, window);
     XGetWindowAttributes(g->display, window, &attr);
     if (XGetTransientForHint(g->display, window, &transient))
@@ -625,13 +609,6 @@ void process_xevent_map(Ghandles * g, XID window)
     write_message(g->vchan, hdr, map_info);
     send_wmname(g, window);
     //      process_xevent_damage(g, window, 0, 0, attr.width, attr.height);
-
-    if (!attr.override_redirect) {
-        /* WM_STATE is always set to normal */
-        new_wm_state[0] = NormalState; /* state */
-        new_wm_state[1] = None;        /* icon */
-        XChangeProperty(g->display, window, g->wm_state, g->wm_state, 32, PropModeReplace, (unsigned char *)new_wm_state, 2);
-    }
 }
 
 void process_xevent_unmap(Ghandles * g, XID window)
@@ -646,7 +623,6 @@ void process_xevent_unmap(Ghandles * g, XID window)
     hdr.untrusted_len = 0;
     write_struct(g->vchan, hdr);
     XDeleteProperty(g->display, window, g->wm_state);
-    XDeleteProperty(g->display, window, g->net_wm_state);
 }
 
 void process_xevent_destroy(Ghandles * g, XID window)
@@ -738,18 +714,18 @@ void process_xevent_configure(Ghandles * g, XID window,
     conf.height = ev->height;
     conf.override_redirect = ev->override_redirect;
     write_message(g->vchan, hdr, conf);
-    send_pixmap_mfns(g, window);
+    send_pixmap_grant_refs(g, window);
 }
 
-void send_clipboard_data(libvchan_t *vchan, char *data, int len)
+void send_clipboard_data(libvchan_t *vchan, XID window, char *data, uint32_t len)
 {
     struct msg_hdr hdr;
     hdr.type = MSG_CLIPBOARD_DATA;
+    hdr.window = window;
     if (len > MAX_CLIPBOARD_SIZE)
+    {
         len = MAX_CLIPBOARD_SIZE;
-    else
-        hdr.window = len;
-    hdr.window = len;
+    }
     hdr.untrusted_len = len;
     write_struct(vchan, hdr);
     write_data(vchan, (char *) data, len);
@@ -820,7 +796,7 @@ void process_xevent_selection(Ghandles * g, XSelectionEvent * ev)
                 Utf8_string_atom, Qprop,
                 g->stub_win, CurrentTime);
     else
-        send_clipboard_data(g->vchan, (char *) data, len);
+        send_clipboard_data(g->vchan, g->stub_win, (char *) data, len);
     /* even if the clipboard owner does not support UTF8 and we requested
        XA_STRING, it is fine - ascii is legal UTF8 */
     XFree(data);
@@ -834,6 +810,8 @@ void process_xevent_selection_req(Ghandles * g,
     Atom Targets = XInternAtom(g->display, "TARGETS", False);
     Atom Compound_text =
         XInternAtom(g->display, "COMPOUND_TEXT", False);
+    Atom Utf8_string_atom =
+        XInternAtom(g->display, "UTF8_STRING", False);
     int convert_style = XConverterNotFound;
 
     if (g->log_level > 0)
@@ -841,7 +819,7 @@ void process_xevent_selection_req(Ghandles * g,
                 XGetAtomName(g->display, req->target));
     resp.property = None;
     if (req->target == Targets) {
-        Atom tmp[4] = { XA_STRING, Targets, g->utf8_string_atom,
+        Atom tmp[4] = { XA_STRING, Targets, Utf8_string_atom,
             Compound_text
         };
         XChangeProperty(g->display, req->requestor, req->property,
@@ -850,34 +828,17 @@ void process_xevent_selection_req(Ghandles * g,
                 tmp, sizeof(tmp) / sizeof(tmp[0]));
         resp.property = req->property;
     }
+    if (req->target == Utf8_string_atom)
+        convert_style = XUTF8StringStyle;
     if (req->target == XA_STRING)
         convert_style = XTextStyle;
-    else if (req->target == Compound_text)
+    if (req->target == Compound_text)
         convert_style = XCompoundTextStyle;
-    else if (req->target == g->utf8_string_atom)
-        convert_style = XUTF8StringStyle;
     if (convert_style != XConverterNotFound) {
         XTextProperty ct;
-        char empty_string[1] = { '\0' };
-        char *ptr[] = { empty_string };
-        // Workaround for an Xlib bug: Xutf8TextListToTextProperty mangles
-        // certain characters, so we check for UTF-8 validity ourselves and use
-        // XStringListToTextProperty.  If we fail a UTF-8 validity check, ptr[0]
-        // is left pointing to an empty string, which is safe.
-        //
-        // We don’t return immediately because users might (reasonably)
-        // assume that any sensitive data previously on the clipboard
-        // (such as passwords) has been overwritten.
-        if (convert_style == XUTF8StringStyle &&
-            !is_valid_clipboard_string_from_vm((unsigned char *)g->clipboard_data))
-            fputs("Invalid clipboard data from VM\n", stderr);
-        else
-            ptr[0] = (char *) g->clipboard_data;
-        if (!XStringListToTextProperty(ptr, 1, &ct)) {
-            fputs("Out of memory in Xutf8TextListToTextProperty()\n", stderr);
-            return;
-        }
-        ct.encoding = req->target;
+        Xutf8TextListToTextProperty(g->display,
+                (char **) &g->clipboard_data,
+                1, convert_style, &ct);
         XSetTextProperty(g->display, req->requestor, &ct,
                 req->property);
         XFree(ct.value);
@@ -1062,7 +1023,7 @@ void process_xevent_message(Ghandles * g, XClientMessageEvent * ev)
                 fprintf(stderr, "unhandled tray opcode: %ld\n",
                         ev->data.l[1]);
         }
-    } else if (ev->message_type == g->net_wm_state) {
+    } else if (ev->message_type == g->wm_state) {
         struct msg_hdr hdr;
         struct msg_window_flags msg;
 
@@ -1211,7 +1172,7 @@ int send_full_window_info(Ghandles *g, XID w, struct window_data *wd)
     conf.height = attr.height;
     conf.override_redirect = attr.override_redirect;
     write_message(g->vchan, hdr, conf);
-    send_pixmap_mfns(g, w);
+    send_pixmap_grant_refs(g, w);
 
     send_wmclass(g, w, 1);
     send_wmnormalhints(g, w, 1);
@@ -1361,8 +1322,7 @@ void mkghandles(Ghandles * g)
     g->tray_opcode =
         XInternAtom(g->display, "_NET_SYSTEM_TRAY_OPCODE", False);
     g->xembed_info = XInternAtom(g->display, "_XEMBED_INFO", False);
-    g->wm_state = XInternAtom(g->display, "WM_STATE", False);
-    g->net_wm_state = XInternAtom(g->display, "_NET_WM_STATE", False);
+    g->wm_state = XInternAtom(g->display, "_NET_WM_STATE", False);
     g->wm_state_fullscreen = XInternAtom(g->display, "_NET_WM_STATE_FULLSCREEN", False);
     g->wm_state_demands_attention = XInternAtom(g->display, "_NET_WM_STATE_DEMANDS_ATTENTION", False);
     g->wm_take_focus = XInternAtom(g->display, "WM_TAKE_FOCUS", False);
@@ -1686,14 +1646,9 @@ void handle_focus(Ghandles * g, XID winid)
     } else if (key.type == FocusOut
             && (key.mode == NotifyNormal
                 || key.mode == NotifyUngrab)) {
-        if ( (l=list_lookup(windows_list, winid)) && (l->data) )
-            input_hint = ((struct window_data*)l->data)->input_hint;
-        else {
-            fprintf(stderr, "WARNING handle_focus: Window 0x%x data not initialized", (int)winid);
-            input_hint = True;
-        }
-        if (input_hint)
-            XSetInputFocus(g->display, None, RevertToParent, CurrentTime);
+
+        XSetInputFocus(g->display, None, RevertToParent,
+                CurrentTime);
 
         if (g->log_level > 1)
             fprintf(stderr, "0x%x lost focus\n", (int) winid);
@@ -1760,43 +1715,26 @@ void handle_map(Ghandles * g, XID winid)
 
 void handle_close(Ghandles * g, XID winid)
 {
-    struct genlist *l;
-    int use_delete_window;
-
-    if ( (l=list_lookup(windows_list, winid)) && (l->data) ) {
-        use_delete_window = ((struct window_data*)l->data)->support_delete_window;
-    } else {
-        fprintf(stderr, "WARNING handle_close: Window 0x%x data not initialized",
-                (int)winid);
-        use_delete_window = True; /* gentler, though it may be a no-op */
-    }
-
-    if (use_delete_window) {
-        XClientMessageEvent ev;
-        memset(&ev, 0, sizeof(ev));
-        ev.type = ClientMessage;
-        ev.display = g->display;
-        ev.window = winid;
-        ev.format = 32;
-        ev.message_type = g->wmProtocols;
-        ev.data.l[0] = g->wmDeleteMessage;
-        //        XSetInputFocus(g->display, winid, RevertToParent, CurrentTime);
-        XSendEvent(ev.display, ev.window, TRUE, 0, (XEvent *) & ev);
-        if (g->log_level > 0)
-            fprintf(stderr, "wmDeleteMessage sent for 0x%x\n",
-                    (int) winid);
-    } else {
-        XKillClient(g->display, winid);
-        if (g->log_level > 0)
-            fprintf(stderr, "XKillClient() called for 0x%x\n",
-                    (int) winid);
-    }
+    XClientMessageEvent ev;
+    memset(&ev, 0, sizeof(ev));
+    ev.type = ClientMessage;
+    ev.display = g->display;
+    ev.window = winid;
+    ev.format = 32;
+    ev.message_type = g->wmProtocols;
+    ev.data.l[0] = g->wmDeleteMessage;
+    //        XSetInputFocus(g->display, winid, RevertToParent, CurrentTime);
+    XSendEvent(ev.display, ev.window, TRUE, 0, (XEvent *) & ev);
+    if (g->log_level > 0)
+        fprintf(stderr, "wmDeleteMessage sent for 0x%x\n",
+                (int) winid);
 }
 
 /* start X server, returns its PID
  */
 pid_t do_execute_xorg(
-        char *w_str, char *h_str, char *mem_str, char *depth_str)
+        char *w_str, char *h_str, char *mem_str, char *depth_str,
+        char *gui_domid_str)
 {
     pid_t pid;
     int fd;
@@ -1811,6 +1749,7 @@ pid_t do_execute_xorg(
             setenv("H", h_str, 1);
             setenv("MEM", mem_str, 1);
             setenv("DEPTH", depth_str, 1);
+            setenv("GUI_DOMID", gui_domid_str, 1);
             /* don't leak other FDs */
             for (fd = 3; fd < 256; fd++)
                 close(fd);
@@ -1833,7 +1772,7 @@ void terminate_and_cleanup_xorg(Ghandles *g) {
 }
 
 #define CLIPBOARD_4WAY
-void handle_clipboard_req(Ghandles * g, XID UNUSED(winid))
+void handle_clipboard_req(Ghandles * g, XID winid)
 {
     Atom Clp;
     Atom QProp = XInternAtom(g->display, "QUBES_SELECTION", False);
@@ -1849,14 +1788,14 @@ void handle_clipboard_req(Ghandles * g, XID UNUSED(winid))
         fprintf(stderr, "clipboard req, owner=0x%x\n",
                 (int) owner);
     if (owner == None) {
-        send_clipboard_data(g->vchan, NULL, 0);
+        send_clipboard_data(g->vchan, winid, NULL, 0);
         return;
     }
     XConvertSelection(g->display, Clp, Targets, QProp,
             g->stub_win, CurrentTime);
 }
 
-void handle_clipboard_data(Ghandles * g, int len)
+void handle_clipboard_data(Ghandles * g, XID UNUSED(winid), unsigned int len)
 {
     Atom Clp = XInternAtom(g->display, "CLIPBOARD", False);
 
@@ -1895,7 +1834,7 @@ void handle_window_flags(Ghandles *g, XID winid)
     read_data(g->vchan, (char *) &msg_flags, sizeof(msg_flags));
 
     /* FIXME: only first 10 elements are parsed */
-    ret = XGetWindowProperty(g->display, winid, g->net_wm_state, 0, 10,
+    ret = XGetWindowProperty(g->display, winid, g->wm_state, 0, 10,
             False, XA_ATOM, &act_type, &act_fmt, &nitems, &bytesleft, (unsigned char**)&state_list);
     if (ret != Success)
         return;
@@ -1928,7 +1867,7 @@ void handle_window_flags(Ghandles *g, XID winid)
     if (!changed)
         return;
 
-    XChangeProperty(g->display, winid, g->net_wm_state, XA_ATOM, 32, PropModeReplace, (unsigned char *)new_state_list, j);
+    XChangeProperty(g->display, winid, g->wm_state, XA_ATOM, 32, PropModeReplace, (unsigned char *)new_state_list, j);
 }
 
 void handle_message(Ghandles * g)
@@ -1967,7 +1906,7 @@ void handle_message(Ghandles * g)
             handle_clipboard_req(g, hdr.window);
             break;
         case MSG_CLIPBOARD_DATA:
-            handle_clipboard_data(g, hdr.window);
+            handle_clipboard_data(g, hdr.window, hdr.untrusted_len);
             break;
         case MSG_KEYMAP_NOTIFY:
             handle_keymap_notify(g);
@@ -1983,17 +1922,18 @@ void handle_message(Ghandles * g)
     }
 }
 
-pid_t get_xconf_and_run_x(libvchan_t *vchan)
+pid_t get_xconf_and_run_x(Ghandles *g)
 {
     struct msg_xconf xconf;
-    char w_str[12], h_str[12], mem_str[12], depth_str[12];
+    char w_str[12], h_str[12], mem_str[12], depth_str[12], gui_domid_str[12];
     pid_t x_pid;
-    read_struct(vchan, xconf);
+    read_struct(g->vchan, xconf);
     snprintf(w_str, sizeof(w_str), "%d", xconf.w);
     snprintf(h_str, sizeof(h_str), "%d", xconf.h);
     snprintf(mem_str, sizeof(mem_str), "%d", xconf.mem);
     snprintf(depth_str, sizeof(depth_str), "%d", xconf.depth);
-    x_pid = do_execute_xorg(w_str, h_str, mem_str, depth_str);
+    snprintf(gui_domid_str, sizeof(gui_domid_str), "%u", g->domid);
+    x_pid = do_execute_xorg(w_str, h_str, mem_str, depth_str, gui_domid_str);
     if (x_pid == (pid_t)-1) {
         errx(1, "X server startup failed");
     }
@@ -2044,6 +1984,7 @@ void usage()
     fprintf(stderr, "       -M  sync only Caps Lock key event\n");
     fprintf(stderr, "       -c  turn off composite \"redirect automatic\" mode\n");
     fprintf(stderr, "       -h  print this message\n");
+    fprintf(stderr, "       -d  GUI domain id (default: 0)\n");
     fprintf(stderr, "\n");
     fprintf(stderr, "Log levels:\n");
     fprintf(stderr, " 0 - only errors\n");
@@ -2059,7 +2000,8 @@ void parse_args(Ghandles * g, int argc, char **argv)
     g->log_level = 0;
     g->sync_all_modifiers = 1;
     g->composite_redirect_automatic = 1;
-    while ((opt = getopt(argc, argv, "qvchmM")) != -1) {
+    g->domid = 0;
+    while ((opt = getopt(argc, argv, "qvchmMd:")) != -1) {
         switch (opt) {
             case 'q':
                 g->log_level--;
@@ -2079,6 +2021,9 @@ void parse_args(Ghandles * g, int argc, char **argv)
             case 'h':
                 usage();
                 exit(0);
+            case 'd':
+                g->domid = atoi(optarg);
+                break;
             default:
                 usage();
                 exit(1);
@@ -2096,8 +2041,9 @@ int main(int argc, char **argv)
     Ghandles g;
     int wait_fds[2];
 
-    /* FIXME: 0 is remote domain */
-    g.vchan = libvchan_server_init(0, 6000, 4096, 4096);
+    parse_args(&g, argc, argv);
+
+    g.vchan = libvchan_server_init(g.domid, 6000, 4096, 4096);
     if (!g.vchan) {
         fprintf(stderr, "vchan initialization failed\n");
         exit(1);
@@ -2108,10 +2054,9 @@ int main(int argc, char **argv)
     saved_argv = argv;
     vchan_register_at_eof(handle_guid_disconnect);
     send_protocol_version(g.vchan);
-    g.x_pid = get_xconf_and_run_x(g.vchan);
+    g.x_pid = get_xconf_and_run_x(&g);
     mkghandles(&g);
     ghandles_for_vchan_reinitialize = &g;
-    parse_args(&g, argc, argv);
     /* Turn on Composite for all children of root window. This way X server
      * keeps separate buffers for each (root child) window.
      * There are two modes:
